using LinearAlgebra

import Random: randperm!, seed!
import StatsBase: knuths_sample!
import SparseArrays: findnz, sparse, SparseMatrixCSC

# Generate a bitvector with `n_activations` entries set to true
function generate_sparse_bitvector(N::Int, n_activations::Int)::BitVector
    p = Array{Int}(undef, n_activations)
    knuths_sample!(1:N, p)
    v = falses(N)
    v[p] .= true
    return v
end

# An HTM layer consists of N minicolumns, each with M cells
# All cells in a minicolumn share the same feedforward dendrite
# An HTM cell can be in one of three states:
#   active, predictive, or inactive
# Each cell has B basal dendrites, each with K potential
# synapses to other cells in the same layer
# Each potential synapse has a permanence value between 0 and 1,
# which is connected if the permanence is above a threshold value
struct Layer
	# A B×M×N array of the bth basal dendrite for the ith cell in the jth minicolumn.
	# Each basal dendrite is a sparse M×N matrix with K permanence values
	# representing the potential synapses to other cells in the layer
	basal_dendrite_permanences::Array{SparseMatrixCSC{Float32, Int}, 3}

	# A B×M×N array of the bth basal dendrite used-ness for the ith cell in the jth minicolumn.
	basal_dendrite_usedness::Array{SparseMatrixCSC{Bool, Int}, 3}

	# The M×N binary matrix that is true if the ith cell in the jth minicolumn
	# is in an active state
	A::BitMatrix

	# The M×N binary matrix that is true if the ith cell in the jth minicolumn
	# is in a predictive state
	Π::BitMatrix
end

function Base.hash(layer::Layer, h::UInt)::UInt
	u1 = hash(layer.basal_dendrite_permanences)
	u2 = hash(layer.basal_dendrite_usedness)
	u3 = hash(layer.A)
	u4 = hash(layer.Π)
	return hash(u1, hash(u2, hash(u3, hash(u4, h))))
end

function create_layer(
	M::Int, # number of cells per minicolumn
	N::Int, # number of minicolumns
	B::Int, # number of basal dendrites per cell
	K::Int, # number of potential synapses per cell
	α::Float32, # permanence threshold for whether a synapse is being used
	)

	# Allocate state
	A = falses(M, N)
	Π = falses(M, N)

	# Allocate basal dendrites
	basal_dendrite_permanences = Array{SparseMatrixCSC{Float32, Int}, 3}(undef, (B,M,N))
	basal_dendrite_usedness = similar(basal_dendrite_permanences)
	row_cols = Vector{Int}(undef, K)
	mn = CartesianIndices((M, N))
	for n in 1:N
		for m in 1:M
			for b in 1:B
				rows = Vector{Int}(undef, K)
				cols = Vector{Int}(undef, K)
				permanences = rand(Float32, K) # uniform random

				valid_sample = false
				while !valid_sample
					knuths_sample!(1:M*N, row_cols)
					for (k,v) in enumerate(row_cols)
						colrow = mn[v]
						rows[k] = colrow[1]
						cols[k] = colrow[2]
					end

					# ensure that we don't have any self-references
					valid_sample = !any(
						row==m && col==n for (row,col) in (rows, cols)
						)
				end

				basal_dendrite_permanences[b,m,n] = sparse(rows, cols, permanences, M, N)
				basal_dendrite_usedness[b,m,n] = sparse(rows, cols, permanences .> α, M, N)
			end
		end
	end

	return Layer(basal_dendrite_permanences, basal_dendrite_usedness, A, Π)
end

function reinforce_dendrite!(
	output::Layer,
	layer::Layer,
	m::Int,
	n::Int,
	b::Int,
	δ⁺::Float32, # positive learning delta (value is positive)
	δ⁻::Float32, # negative learning delta (value is positive)
	)
	A = layer.A
	D = output.basal_dendrite_permanences[b,m,n]

	# apply decay to all synapses
	D.nzval[:] .-= δ⁻

	# increase permanences of active presynaptic cells
	for (k,(i,j,u)) in enumerate(zip(findnz(D)...))
		D.nzval[k] += δ⁺ * A[i,j]
	end

	return output
end

"""
Pass in a vector of feedforward activations (true if minicolumn is active)
The ff is of length N
Mutates the subsequent later to produce the next state
"""
function update!(
	output::Layer,
	layer::Layer,
	ff::BitVector,
	α::Float32,  # permanence threshold
	θ::Int,      # dendrite activation threshold (number of synapses that must be active to activate the dendrite)
	δ⁺::Float32, # positive learning delta (value is positive)
	δ⁻::Float32, # negative learning delta (value is positive)
	)
	B, M, N = size(layer.basal_dendrite_permanences)

	# Initialize next state to inactive
	fill!(output.A, false)
	fill!(output.Π, false)

	# Copy over the permanence values
	for n in 1:N
		for m in 1:M
			for b in 1:B
				output.basal_dendrite_permanences[b,m,n].nzval[:] = layer.basal_dendrite_permanences[b,m,n].nzval
				output.basal_dendrite_usedness[b,m,n].nzval[:] = layer.basal_dendrite_usedness[b,m,n].nzval
			end
		end
	end

	# A cell becomes active if its ff dendrite is active and
	# it was either predicted or no cell in its column was predicted
	for (n, column_activated) in enumerate(ff)
		if column_activated
			# input is active
			n_predicted_cells = 0
			for m in 1:M
				if layer.Π[m,n] # whether it was in a predicted state
					n_predicted_cells += 1
					output.A[m,n] = true
				end
			end

			if n_predicted_cells == 0
				# activate all cells in the mini-column
				output.A[:,n] .= true
			end
		end
	end

	# A cell becomes predictive if any of its basal dendrites have enough active synapses
	# TODO: For now, a cell can be both predictive and active
	for n in 1:N
		for m in 1:M
			if any(layer.basal_dendrite_usedness[b,m,n] ⋅ output.A ≥ θ for b in 1:B)
				output.Π[m,n] = true
			end
		end
	end

	# Apply reinforcement learning rule
	# A cell that went from predicted -> active gets reinforced
	# A cell that went from inactive -> active for which no cell in its column was predicted causes us to reinforce
	for (n, column_activated) in enumerate(ff)
		if column_activated
			if !any(layer.Π[:,n])
				# no cell was predicted
				# pick the cell closest to being active and reinforce its closest-to-active dendrite
				best = (1,1,0) # (m,b,num_active_synapses)
				for m in 1:M
					(v,b) = findmax(b -> begin
						tot = 0
						for (i,j,u) in zip(findnz(layer.basal_dendrite_usedness[b,m,n])...)
							tot += layer.A[i,j]
						end
						return tot
					end, 1:B)

					if v > best[3] || (v == best[3] && rand() > 0.5)
						best = (m,b,v)
					end
				end

				# reinforce this cell's dendrite
				m = best[1]
				b = best[2]
				reinforce_dendrite!(output, layer, m, n, b, δ⁺, δ⁻)
			end

			for m in 1:M
				if output.A[m,n]
					# cell became active
					if layer.Π[m,n]
						# was predicted - reinforce any active dendrites
						for b in 1:B
							if layer.basal_dendrite_usedness[b,m,n] ⋅ layer.A ≥ θ
								reinforce_dendrite!(output, layer, m, n, b, δ⁺, δ⁻)
							end
						end
					end
				end

				# ensure we are clamped and keep the usedness up to date
				for b in 1:B
					clamp!(output.basal_dendrite_permanences[b,m,n], 0.0, 1.0)
					output.basal_dendrite_usedness[b,m,n].nzval[:] = layer.basal_dendrite_permanences[b,m,n].nzval .> α
				end
			end
		else
			# Column was not activated
			# we also apply a small decay to active dendrites of cells that were predicted but did not become active
			for m in 1:M
				if layer.Π[m,n] # whether it was in a predicted state
					for b in 1:B
						if layer.basal_dendrite_usedness[b,m,n] ⋅ layer.A ≥ θ
							output.basal_dendrite_permanences[b,m,n].nzval[:] .-= δ⁻/2
						end
					end
				end
			end
		end
	end

	return output
end

function calc_num_full_column_activations(layer::Layer)
	B, M, N = size(layer.basal_dendrite_permanences)
	n_full_column_activations = 0
	for n in 1:N
		n_full_column_activations += sum(layer.A[:,n]) == M
	end
	return n_full_column_activations
end

function calc_num_full_column_activations_in(layer::Layer, target)
	B, M, N = size(layer.basal_dendrite_permanences)
	n_full_column_activations = 0
	for n in 1:N
		if n ∈ target && sum(layer.A[:,n]) == M
			n_full_column_activations += 1
		end
	end
	return n_full_column_activations
end

calc_num_active_cells(layer::Layer)::Int = sum(layer.A)
calc_num_predictive_cells(layer::Layer)::Int = sum(layer.Π)
calc_num_used_synapses(layer)::Int = sum(sum(U) for U in layer.basal_dendrite_usedness)
calc_num_saturated_permanences(layer)::Int = sum(P -> sum(p == 0 || p == 1 for p in P.nzval), layer.basal_dendrite_permanences)
calc_total_permanence(layer)::Float32 = sum(sum(P) for P in layer.basal_dendrite_permanences)

function calc_columnwise_max!(π::BitVector, Π::BitMatrix)
	M, N = size(Π)
	for n in 1:N
		columnwise_max = false
		for m in 1:M
			columnwise_max = max(columnwise_max, Π[m,n])
		end
		π[n] = columnwise_max
	end
	return π
end

"""
Compute the prediction accuracy of the given ff input given a layer.
This method assumes that our basal dendrites have set the predictive state of the layer,
and that the layer predicts the next input based on having cells in a predictive state in
the mini-columns corresponding to the set bits in the next ff input.
We use a softmax prediction distribution: P(token) ∝ exp(β (ff ⋅ π)),
where β is a scalar and π is the columnwise max of Π, and thus (ff ⋅ π) is the number of
minicolumns that have at least one cell in a predictive state.

We set `weights` to the exp(β (ff ⋅ π)) values for each token,
and we set `π` in each call as well.
"""
function predict!(
	weights::Vector{Float32},  # of length |tokens|
	π::BitVector,              # of length N
	tokens::Vector{BitVector}, # each token has length N
	layer::Layer,
	β::Float32
	)

	calc_columnwise_max!(π, layer.Π)
	for (i_token, token) in enumerate(tokens)
		weights[i_token] = exp(β*(π ⋅ token))
	end
	return weights
end
